#!/usr/bin/env python
# coding: utf-8

__version__='0.1.2'
import torch

# -------------------------------------------------------
def _get_module_type(module):
	return module.__class__.__name__
def _is_support_layer(module):
	return _get_module_type(module) in {'Linear', 'Conv2d','BatchNorm2d','BatchNorm1d'}

# -------------------------------------------------------
def _compute_grad1_for_linear(module, Fp, Bp):
	module.weight.grad1 = torch.einsum('ni,nj->nij', Bp, Fp)
	if module.bias is not None:
		module.bias.grad1 = Bp

def _compute_grad1_for_conv2d(module, Fp, Bp):
	n = Fp.shape[0]
	Fp = torch.nn.functional.unfold(Fp, module.kernel_size, module.dilation, module.padding, module.stride)
	Bp = Bp.reshape(n, -1, Fp.shape[-1])
	grad1 = torch.einsum('ijk,ilk->ijl', Bp, Fp)
	shape = [n] + list(module.weight.shape)
	grad1 = grad1.reshape(shape)
	module.weight.grad1 = grad1
	if module.bias is not None:
		module.bias.grad1 = torch.sum(Bp, dim=2)

def _compute_grad1_for_batchnorm2d(module, Fp, Bp):
	module.weight.grad1 = torch.sum(Fp*Bp, dim=[2,3])
	if module.bias is not None:
		module.bias.grad1 = torch.sum(Bp, dim=[2,3])

def _compute_grad1_for_batchnorm1d(module, Fp, Bp):
	module.weight.grad1 = Fp*Bp
	if module.bias is not None:
		module.bias.grad1 = Bp

def _set_grad1(module):
	module_type = _get_module_type(module)
	Fp = module._forward1
	Bp = module._backward1

	if module_type=='Linear':
		_compute_grad1_for_linear(module, Fp, Bp)
	if module_type=='Conv2d':
		_compute_grad1_for_conv2d(module, Fp, Bp)
	if module_type=='BatchNorm2d':
		_compute_grad1_for_batchnorm2d(module, Fp, Bp)
	if module_type=='BatchNorm1d':
		_compute_grad1_for_batchnorm1d(module, Fp, Bp)

# -------------------------------------------------------
def _capture_forwardprops(module, inputs, _outputs):
	module._forward1	= inputs[0].detach()

def _capture_backwardprops(module, _grad_inputs, grad_outputs):
	module._backward1	= grad_outputs[0].detach()
	assert hasattr(module, '_forward1'), 'please add_hooks before forward-propagation'
	_set_grad1(module)

def register(model):
	handles = []
	for module in model.modules():
		if _is_support_layer(module):
			handles.append(module.register_forward_hook(_capture_forwardprops))
			handles.append(module.register_full_backward_hook(_capture_backwardprops))
	model.__dict__.setdefault('autograd_hacks_hooks', []).extend(handles)

# -------------------------------------------------------
def deregister(model):
	for module in model.modules():
		if hasattr(module, '_forward1'):
			del module._forward1
		if hasattr(module, '_backward1'):
			del module._backward1
	for param in model.parameters():
		if hasattr(param, 'grad1'):
			del param.grad1
	if hasattr(model, 'autograd_hacks_hooks'):
		for handle in model.autograd_hacks_hooks:
			handle.remove()
		del model.autograd_hacks_hooks

# -------------------------------------------------------




#!/usr/bin/env python
# coding: utf-8

# -------------------------------------------------------
import torch
import torch.nn.functional as F
import exgrads.hooks as hooks
def vectorize(model, x, y):
	model.eval()
	hooks.register(model)

	logit = model(x)
	loss  = F.cross_entropy(logit, y, reduction='sum')
	model.zero_grad()
	loss.backward()

	vectors = []
	for param in model.parameters():
		if not hasattr(param, 'grad1'): continue
		grad1 = param.grad1.detach()
		grad1 = grad1.flatten(1,-1)
		vectors.append(grad1)
		del param.grad1
	vectors = torch.cat(vectors,dim=1)

	hooks.deregister(model)
	return vectors


# -------------------------------------------------------
import torch
import torch.nn.functional as F
import exgrads.hooks as hooks
def normL2(model, x, y):
	model.eval()
	hooks.register(model)

	logit = model(x)
	loss  = F.cross_entropy(logit, y, reduction='sum')
	model.zero_grad()
	loss.backward()

	norms = torch.zeros(x.shape[0])
	norms = norms.to(x.device)

	for param in model.parameters():
		if not hasattr(param, 'grad1'): continue
		grad1 = param.grad1.detach()
		grad1 = grad1.flatten(1,-1)
		norms += grad1.pow(2).sum(dim=1)
		del param.grad1
	norms = norms.pow(0.5)
	
	hooks.deregister(model)
	return norms





# -------------------------------------------------------




#!/usr/bin/env python
# coding: utf-8

# -------------------------------------------------------
import torch
import exgrads.hooks as hooks
def trH(model, x):
	model.eval()
	hooks.register(model)
	
	results = torch.zeros(x.shape[0])		# (b)
	results = results.to(x.device)

	logit = model(x)								# := (b,K)

	# for p*p^T part
	Z		= torch.logsumexp(logit, dim=1)			# -> (b)
	loss 	= Z.sum()								# -> (1)
	
	loss.backward(retain_graph=True)
	model.zero_grad()
	for param in model.parameters():
		grad1 = param.grad1**2 						# := (b,*)
		grad1 = grad1.flatten(1,-1) 				# -> (b,prod(*))
		grad1 = torch.sum(grad1,dim=1) 				# -> (b,)
		results -= grad1.detach()
		del param.grad1

	# for diag(p) part
	prob	= torch.softmax(logit,dim=1)				# -> (b,K)
	prob	= prob.unsqueeze(dim=2)						# -> (b,K,1)
	logit	= logit.sum(dim=0)							# -> (K)
	for k, logit_k in enumerate(logit):
		logit_k.backward(retain_graph=True)
		model.zero_grad()
		for param in model.parameters():
			grad1 = param.grad1**2					# := (b,*)
			grad1 = grad1.flatten(1,-1)				# -> (b,prod(*))
			grad1 = prob[:,k] * grad1				# -> (b,prod(*))
			grad1 = grad1.flatten(1,-1)				# -> (b)
			grad1 = torch.sum(grad1,dim=1)			# -> (b)
			results += grad1.detach()
			del param.grad1
	
	hooks.deregister(model)
	return results


# -------------------------------------------------------




from setuptools import setup
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.md'), encoding='utf-8') as fp:
	long_description = fp.read()

setup(
	name='ExGrads',
	packages=['exgrads'],
	version='0.1.7',
	license='MIT',
	
	install_requires=['torch'],

	author='Takuo Hamaguchi',
	author_email='nyahha@gmail.com',
	
	url='https://gitlab.com/takuo-h/examplewise-gradients',

	description='calculate example-wise gradient',
	long_description=long_description,
	long_description_content_type='text/markdown',
	keywords='',

	classifiers=[
		'License :: OSI Approved :: MIT License',
		'Programming Language :: Python :: 3.6',
	],
)